﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System;
  using Common;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  /// <summary>
  ///     Pushes the sum of the top two items.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.+")]
  public class FloatAddExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.FloatStack.Count < 2)
        return false;

      var first = interpreter.FloatStack.Pop();
      var second = interpreter.FloatStack.Top;
      var result = second + first;

      if (double.IsPositiveInfinity(result)) result = double.MaxValue;
      if (double.IsNegativeInfinity(result)) result = double.MinValue;

      interpreter.FloatStack.SetTop(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes the difference of the top two items; that is, the second item minus the top item.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.-")]
  public class FloatSubtractExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.FloatStack.Count < 2)
        return false;

      var first = interpreter.FloatStack.Pop();
      var second = interpreter.FloatStack.Top;
      var result = second - first;

      if (double.IsPositiveInfinity(result)) result = double.MaxValue;
      if (double.IsNegativeInfinity(result)) result = double.MinValue;

      interpreter.FloatStack.SetTop(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes the product of the top two items.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.*")]
  public class FloatMultiplyExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.FloatStack.Count < 2)
        return false;

      var first = interpreter.FloatStack.Pop();
      var second = interpreter.FloatStack.Top;
      var result = second * first;

      if (double.IsPositiveInfinity(result)) result = double.MaxValue;
      if (double.IsNegativeInfinity(result)) result = double.MinValue;

      interpreter.FloatStack.SetTop(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes the quotient of the top two items; that is, the second item divided by the top item.
  ///     If the top item is zero this acts as a NOOP.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT./")]
  public class FloatDivideExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.FloatStack.Count < 2 ||
          interpreter.FloatStack.Top.IsAlmost(0))
        return false;

      var first = interpreter.FloatStack.Pop();
      var second = interpreter.FloatStack.Top;
      var result = second / first;

      if (double.IsPositiveInfinity(result)) result = double.MaxValue;
      if (double.IsNegativeInfinity(result)) result = double.MinValue;

      interpreter.FloatStack.SetTop(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes the second stack item modulo the top stack item. If the top item is zero this acts as a NOOP.
  ///     The modulus is computed as the remainder of the quotient, where the quotient has first been truncated toward
  ///     negative infinity.
  ///     (This is taken from the definition for the generic MOD function in Common Lisp, which is described
  ///     for example at http://www.lispworks.com/reference/HyperSpec/Body/f_mod_r.htm.)
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.%")]
  public class FloatModuloExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.FloatStack.Count < 2 ||
          interpreter.FloatStack.Top.IsAlmost(0))
        return false;

      var first = interpreter.FloatStack.Pop();
      var second = interpreter.FloatStack.Top;
      var result = second % first;

      interpreter.FloatStack.SetTop(result);
      return true; ;
    }
  }

  /// <summary>
  ///     Pushes the minimum of the top two items.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.MIN")]
  public class FloatMinExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.FloatStack.Count < 2)
        return false;

      var first = interpreter.FloatStack.Pop();
      var second = interpreter.FloatStack.Top;
      var result = Math.Min(second, first);

      interpreter.FloatStack.SetTop(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes the maximum of the top two items.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.MAX")]
  public class FloatMaxExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.FloatStack.Count < 2)
        return false;

      var first = interpreter.FloatStack.Pop();
      var second = interpreter.FloatStack.Top;
      var result = Math.Max(second, first);

      interpreter.FloatStack.SetTop(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the second item is less than the top item, or FALSE otherwise.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.<", StackTypes.Boolean)]
  public class FloatSmallerThanExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.FloatStack.Count < 2)
        return false;

      var first = interpreter.FloatStack.Top;
      var second = interpreter.FloatStack[1];
      interpreter.FloatStack.Remove(2);

      var result = second < first;
      interpreter.BooleanStack.Push(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the second item is less than the top item, or FALSE otherwise.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.<=", StackTypes.Boolean)]
  public class FloatSmallerThanOrEqualExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.FloatStack.Count < 2)
        return false;

      var first = interpreter.FloatStack.Top;
      var second = interpreter.FloatStack[1];
      interpreter.FloatStack.Remove(2);

      var result = second < first || second.IsAlmost(first);
      interpreter.BooleanStack.Push(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the second item is greater than the top item, or FALSE otherwise.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.>", StackTypes.Boolean)]
  public class FloatGreaterThanExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.FloatStack.Count < 2)
        return false;

      var first = interpreter.FloatStack.Top;
      var second = interpreter.FloatStack[1];
      interpreter.FloatStack.Remove(2);

      var result = second > first;
      interpreter.BooleanStack.Push(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the second item is greater than or equal to the top item, or FALSE otherwise.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.>=", StackTypes.Boolean)]
  public class FloatGreaterThanOrEqualExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.FloatStack.Count < 2)
        return false;

      var first = interpreter.FloatStack.Top;
      var second = interpreter.FloatStack[1];
      interpreter.FloatStack.Remove(2);

      var result = second > first || second.IsAlmost(first);
      interpreter.BooleanStack.Push(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes the sine of the top item.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.SIN")]
  public class FloatSineExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.FloatStack.IsEmpty)
        return false;

      var result = Math.Sin(interpreter.FloatStack.Top);
      interpreter.FloatStack.SetTop(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes the cosine of the top item.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.COS")]
  public class FloatCosineExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.FloatStack.IsEmpty)
        return false;

      var result = Math.Cos(interpreter.FloatStack.Top);
      interpreter.FloatStack.SetTop(result);
      return true;
    }
  }

  /// <summary>
  ///     Pushes 1 if the top BOOLEAN is TRUE, or 0 if the top BOOLEAN is FALSE.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.FROMBOOLEAN", StackTypes.Boolean)]
  public class FloatFromBooleanExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.BooleanStack.Count == 0) return false;

      var condition = interpreter.BooleanStack.Pop();
      var value = condition ? 1 : 0;

      interpreter.FloatStack.Push(value);

      return true;
    }
  }

  /// <summary>
  ///     Pushes a floating point version of the top INTEGER.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.FROMINTEGER", StackTypes.Integer)]
  public class FloatFromIntegerExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.IntegerStack.Count == 0) return false;

      var value = (double)interpreter.IntegerStack.Pop();

      interpreter.FloatStack.Push(value);

      return true;
    }
  }
}